# =====================================================================
# csa-soap2rpc.awk: SOAP to XML-RPC converter.
#
# Copyright (c) 2007 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# Notes:
#
# This program makes extensive use of XML-RPC <struct> elements on
# output. An input SOAP construct in the form of:
#
#  <Categories>
#   <Cat>cat1</Cat>
#   <Cat>cat2</Cat>
#  </Categories>
#
# will be turned by this script into the XML-RPC fragment:
#
#  ...
#  <member>
#   <name>Categories</name>
#   <value><struct>
#    <member>
#     <name>Cat</name>
#     <value>cat1</value>
#    </member>
#    <member>
#     <name>Cat</name>
#     <value>cat2</value>
#    </member>
#   </struct></value>
#  </member>
# ...
#
# while a better representation would naturally be to have "Cat"
# turned into an array of values "cat1" and "cat2":
#
#  ...
#  <member>
#   <name>Categories</name>
#   <value><struct>
#    <member>
#     <name>Cat</name>
#     <value><array>
#      <data>
#       <value>cat1</value>
#       <value>cat2</value>
#      </data>
#     </array></value>
#    </member>
#   </struct></value>
#  </member>
# ...
#
# Unfortunately, detecting when arrays should be used instead of structs
# would require that the whole input be loaded in storage beforehand and
# parsed as appropriate, possibly recursively; a big waste of resources
# compared with the current SAX-like approach taken by this script. Short
# of an easy solution, we must pass the trouble up to the application layer,
# after having observed that the parameter processing layer of CSA turns
# the above two fragments respectively into:
#
# WWW_2_1_1_N = Cat
# WWW_2_1_2_N = Cat
# WWW_2_1_1 = cat1
# WWW_2_1_2 = cat2
# 
# and:
#
# WWW_2_1_1_N = Cat
# WWW_2_1_1_1 = cat1
# WWW_2_1_1_2 = cat2
#
# The receiving method of the SOAP call will then have to apply the rule
# that wherever an input argument is supposed to contain a struct element
# which value is a list (i.e. an array) of values, i.e.:
#
# struct{{name1,array{value1,value2,value3, ...}},{name2,value4},...}
#
# the method will instead have to look for a struct of {name,value} members,
# with all the members having the same name (albeit pointed to by different
# variables), i.e.:
#
# struct{{name1,value1},{name1,value2},{name1,value3},{name2,value4},...}
#
# As always, each value can in turn be a complex XML-RPC piece of data.
#
# That is, for the sake of further clarity:
#
# struct{{$WWW_2_1_1_N,array{$WWW_2_1_1_1,$WWW_2_1_1_2}}}
#
# becomes:
#
# struct{{$WWW_2_1_1_N,$WWW_2_1_1},{$WWW_2_1_2_N,$WWW_2_1_2}}
#
# with $WWW_2_1_1_N and $WWW_2_1_2_N being both equal to 'Cat'.
#
# Note that I have used the notation '$WWW_xxx' in those places where
# I'm referring to the *content* of variables as opposed to their names.
#
# Preliminary tests against this program done with SOAP::Lite/Perl/0.69
# (SOAPsh command-line interface) have shown that things seem to work.
# At the moment this script does not yet support element values passed
# through element attibutes, like "<Element attr1=value1" />", but see
# the relevant notes further down.
# =====================================================================

BEGIN { RS = "<" }

# Note: this script already skips comments even without the next
# line but leaving the latter in place won't hurt anyway.

/^!-- [^>]*-->/ { next }		# skip embedded comments.


/^(soap(12)?|SOAP-ENV):Envelope.*>/ {
  print "<params>"
  next
}

/^(\/soap(12)?|SOAP-ENV):Envelope>/ {
  print "</params>"
  next
}

/^(soap(12)?|SOAP-ENV):Header>/ {
  # The header is optional, so I must account for its absence
  # and insert a dummy '<param>' if necessary.
  header = 1
}

/^(soap(12)?|SOAP-ENV):Body>/ && !header {
  print "<param></param>"		# print a dummy 1st param.
}

/^(soap(12)?|SOAP-ENV):(Header|Body)>/ {
  print "<param><value><struct>"
  next
}

/^(\/soap(12)?|SOAP-ENV):(Header|Body)>/ {
  print "</struct></value></param>"
  next
}

/^\// {

  if (name != "") {
     print "<member><name>" name "</name><value>" value "</value></member>"
     name = ""
  }

  else print "</struct></value></member>"

  next
}

# Note that I accept dots in the method name, although they are not
# allowen in XML element names, but I want to be able to call methods
# with names such as 'blogger.getUsersBlogs' and things like that.
# Regarding datatypes, I will consider just the only two ones that
# really matter: strings and base64. Unless a piece of data is
# explicitly (and hopefully) labeled by the client as being base64,
# it will default to a string (i.e. to the XML-RPC default type).

/^[-a-zA-Z0-9:.]+[^>]*>/ {

  type = tolower($0)
  gsub(/(^[-a-zA-Z0-9:.]+[ \n\t\r]+|>.*)/,"",type)
  if (type ~ /(soap-enc|xsd):base64/) {
     tstart = "<base64>"
     tend = "</base64>"
  }
  else tstart = tend = ""

  if (name != "") tmp = "member><name>" name "</name><value><struct>"
  i = index($0,">")
  name = substr($0,1,i)
  sub(/[^-a-zA-Z0-9:.].*/,"",name)
  value = tstart substr($0,i+1) tend
  if (tmp != "") {
     print "<" tmp
     tmp = ""
  }
}

# Detect empty elements like "<Element [attribute(s)] />"

/\/[ \t]*>/ {

  if (name != "") {

     # If needed, insert the code here to extract empty element
     # values from the element attribute(s) (may be tricky though).
     #
     #value = $0
     # ...

     # Strip any spurious CDATA.
     if (value ~ /^ *$/) value = ""

     print "<member><name>" name \
			"</name><value>" value "</value></member>"
     name = ""
  }

  else print "</struct></value></member>"
}

# =====================================================================
# End of csa-soap2rpc.awk
# =====================================================================
