# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2006,2007 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# void _mrwresponse([string root, string child, string about,\
#	    string struct, string property, string name, string value,\
#	    array aname, array avalue, boolean noauto, string file,\
#	    boolean noenc])
#
# Experimental Machine-readable Web (Semantic Web) response processor.
#
# This is a fist attempt to provide CSA with the capability of
# generating responses that are readable according to the rules of the
# so-called Semantic Web. This facility is still very experimental and
# unstable, so it may change a lot in the future or it may even be
# dropped. Here are some tips on how the underlying logic works, as
# it may look rather obscure ad a first glance. This program has proven
# flexible enough to handle the creation of rather complex and nested
# structures, like
# http://linuxgazette.net/105/misc/oregan/foaf-example.rdf.txt
#
# Here are the rules adopted, with respect to the above example:
#
# <foaf:Person> is a "root" element.
# <foaf:name>Jimmy O'Regan</foaf:name> is a name/value pair.
# <foaf:based_near> is a "child" element.
# <foaf:Point> is a "property" element.
# <geo:lat>52.6796</geo:lat> is a name/value pair.
#
# * Property elements can only appare inside RDF structs (Seq/Alt/Bag).
# * Root and property elements may take an "about" attribute, while 
#   child and name elements may not.
# * Property elements may only appare inside child nodes, which implies
#   that child nodes require an RDF struct.
# * Attributes are accepted only for name elements, and only if the
#   associated value is an URI.
# * Root elements may have name/value pairs, but not property elements
#   neither RDF structs, unless these are incapsulated inside child
#   elements.
# * A child element may not have name/value pairs, unless these are
#   incapsulated inside a struct of properties.
#
# If the above rules are violated, the program will ignore the failing
# arguments and apply sensibly the correct ones, in order for the output
# to parse as valid XML/RDF. So if you wander why the output does not
# look like what you expected, either there's a bug in this program or
# you have got these rules wrong.
#
# This program looks flexible enough to allow for the the generation of
# arbitrarily complex XML structures, although I have not yet tested it
# extensively in that respect. If that were the case then it should be
# possible to produce entirely programmatic responses without requiring
# an external template, or requiring only a very minimal one.
# =====================================================================

function _mrwresponse(root,child,about,struct,property,name,value,\
		      aname,avalue,noauto,file,noenc,		a) {

  # set defaults.
  if (file == "") file = ENVIRON["CSA_RDF_MRW"]

  # make sure file is appended to, or printing multiple MRW
  # response blocks with different ontologies will not be possible!

  printf(_NULL) >> file

  if (_isuri(tolower(about)) == _TRUE || about ~ /^#./)
		about = " rdf:about=\"" _xmlencode(about) "\""
  else if (about == ".") about = " rdf:about=\"\""
  else about = ""

  if (struct == "Alt") struct = "rdf:Alt"
  else if (struct == "Bag") struct = "rdf:Bag"
  else struct = "rdf:Seq"				# default.

  if (root != "") {					# new root ?

     # close previous tags if necessary.

     if (_MRW["property"] != "")
	printf("</%s>\n</rdf:li>\n",_MRW["property"]) > file

     if (_MRW["struct"] != "") printf("</%s>\n",_MRW["struct"]) > file

     if (_MRW["child"] != "") printf("</%s>\n",_MRW["child"]) > file

     if (_MRW["root"] != "") printf("</%s>\n",_MRW["root"]) > file

     # clear old values.
     _MRW["child"] = _MRW["struct"] = _MRW["property"] = ""

     # the fake root value "." simply closes the MRW output.

     if (root == ".") {
	close(file)
	_MRW["root"] = _NULL		# this may be tested elsewhere.
	return
     }

     # set new root.
     printf("<%s%s>\n",root,about) > file; _MRW["root"] = root
  }

  if (child != "") {					# new child ?

     # close previous tags if necessary.

     if (_MRW["property"] != "")
	printf("</%s>\n</rdf:li>\n",_MRW["property"]) > file

     if (_MRW["struct"] != "") printf("</%s>\n",_MRW["struct"]) > file

     if (_MRW["child"] != "") printf("</%s>\n",_MRW["child"]) > file

     # clear old values.
     _MRW["struct"] = _MRW["property"] = ""

     # the fake child value "." simply closes the current child.
     if (child == ".") return

     # set new child.
     printf("<%s>\n<%s>\n",child,struct) > file

     _MRW["child"] = child; _MRW["struct"] = struct
  }

  if (property != "" && _MRW["child"] != "") {		# new property ?

     # Properties can only occur inside childs (use default namespace).
     if (_MRW["child"] == "") _mrwresponse("","csa:Child")

     # close previous tags if necessary.

     if (_MRW["property"] != "")
		printf("</%s>\n</rdf:li>\n",_MRW["property"]) > file

     # the fake property value "." simply closes the current property.
     if (property == ".") return

     # set new property.

     printf("<rdf:li>\n<%s%s>\n",property,about) > file

     _MRW["property"] = property
  }

  if (name != "") {

     if (!noauto && _isuri(tolower(value)) == _TRUE) {

	printf("<%s",name) > file

	# attributes are taken into consideration only for URI-type values.
	for (a in aname)
	    printf(" %s=\"%s\"",aname[a],\
			_xmlencode(avalue[aname[a]])) > file

	printf(" rdf:resource=\"%s\" />\n",_xmlencode(value)) > file
     }

     else {
	if (noenc) printf("<%s>%s</%s>\n",name,value,name) > file
	else printf("<%s>%s</%s>\n",name,_xmlencode(value),name) > file
     }
  }

  # now that the array contents have been consumed I clear them,
  # as otherwise they would need to be cleared by the calling
  # program, or they could be attached also to other values they
  # do not apply to.

  delete aname; delete avalue
}

